"""
    CONFIG UTILITIES
        - classes that parse your config file, build a query iterable and yield queries
"""
import numpy as np
import pandas as pd
import yaml


def _prepare_chunked_dates(
        start_date: str,
        end_date: str,
        day_freq: int
):
    """
        Split up dates in to multiple startDate, endDate pairs which are dayFreq apart
        args:
            start_date (str) : startDate of date range
            end_date (str) : endDate of date range
            day_freq (int)  : gap between startDate, endDate within each pairs
        return:
            dateRanges (list of dict)
    """
    dates = pd.Series(pd.date_range(start=start_date, end=end_date))
    date_freq = dates.groupby(np.arange(len(dates)) // day_freq).agg(
        ['first', 'last'])
    return date_freq.rename(
        columns={
            'first': 'startDate',
            'last': 'endDate'}).astype(str).to_dict('records')


def _restructure_date_ranges(config):
    """
        Restructure dateRanges in to multiple pairs gapped by dayFreq
        args:
            config (dict): config file containing dayFreq, dateRanges keys
        return:
            config (dict)
    """
    # dynamic get kwargs or else set to default
    day_freq = config.pop('dayFreq', '1d')
    start_date = config.get('dateRanges', [])[0].get('startDate', None)
    end_date = config.get('dateRanges', [])[0].get('endDate', None)
    # restructure dateRanges
    date_ranges = _prepare_chunked_dates(start_date, end_date, day_freq)
    # update dateRanges in config
    config.update({'dateRanges': date_ranges})
    return config


# pylint: disable=invalid-name, unused-argument
def _generate_queries(
        viewIDs: list,
        queryConfig: list,
        dateRanges: list,
        pageSize: int = 10000,
        samplingLevel: str = 'LARGE',
        **kwargs
):
    """
    Use the Analytics Service Object to query the
    Google Analytics Reporting API V4.
    args:
        viewIDs (list): list of google analytics view ids
        queryConfig (list): Google query config containing metrics, dimensions, etc.
        dateRanges (list): set of start & end dates.
        pageSize (int): (default: 10000) number of samples per pagination step
        samplingLevel (str): (default:LARGE) sampling level of the query
    yields:
        query generator of queries to run
    """
    # static variables
    for vid in viewIDs:
        for _, single_query in enumerate(queryConfig):
            for date_range in dateRanges:
                print("creating a [fields] query")
                fields = {
                    'viewId': vid,
                    'dateRanges': date_range,
                    **single_query,
                    "samplingLevel": samplingLevel,
                    "pageSize": pageSize}
                # building similar queries per vid
                yield [fields]


def _parser(file_name):
    """Parse config in to dictionary"""
    # load config
    with open(file_name, 'r') as file_:
        return yaml.safe_load(file_)


class GAConfigManager:
    """
        Class that parses config .yaml for queries, reader, writer
        Uses these configs to build required query objects.
    """

    def __init__(self, file_name):
        # load config
        self.config = _parser(file_name=file_name)
        if self.config.pop('chunk_date_range', False):
            self.config = _restructure_date_ranges(self.config)
        # get query generator
        self.query_gen, self.additional_args = self._builder(self.config)

    def _builder(self, config):
        """Config builder method to generate query configs"""
        additional_args = {}
        # get metadata
        metadata = {
            'dateRanges': self.config.get('dateRanges', None),
            'samplingLevel': self.config.get('samplingLevel', None),
            'pageSize': self.config.get('pageSize', None)
        }
        additional_args.update({'metadata': metadata})
        return _generate_queries(**config), additional_args

    def get_query(self):
        """Get a Generator for the queries to be executed"""
        return self.query_gen

    def get_additional_args(self):
        """Getter method to get additional arguments"""
        return self.additional_args



"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import json
from sailthru import SailthruClient


class CustomSailThruReader():
    """
        Custom SailThru Reader
    """

    # pylint: disable=super-init-not-called
    def __init__(self, api_keys_path: str):

        self.sailthru_client = self.auth_credentials(api_keys_path)

    @staticmethod
    def auth_credentials(
            api_keys_path
    ):
        """
        Create a client to communicate with Sailthru
        Args:
            api_keys_path: The file path to the api key and secret.
        Returns:
            A service that is connected to the specified API.
        """
        with open(api_keys_path) as json_file:
            data = json.load(json_file)

        return SailthruClient(data['api_key'], data['api_secret'])

    def call_api(self, query: dict):
        """
            returns results of API call
        """
        response = self.sailthru_client.api_get(query.get('endpoint'), query.get('parameters')[0])
        if not response.is_ok():
            error = response.get_error()
            raise Exception("APIError: {}, Status Code: {}, Error Code: {}".format(
                error.get_message(),
                str(response.get_status_code()),
                str(error.get_error_code())))
        else:
            return response.get_body()



"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import gzip
import io
import json
import os

import boto3
from interface import implements
from sdc_dp_helpers.base_writer import BaseWriter, CustomLocalJsonWriter


class CustomS3GZJsonWriter(CustomLocalJsonWriter, implements(BaseWriter)):
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(
            self,
            file_name: str,
            folder_path: str,
            bucket: str,
            profile_name: str = None,
            **kwargs
    ):
        self.os_path_sep = "/"

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.bucket = bucket
        """Writes a general object to s3"""
        self.s3_resource = self.boto3_session.resource('s3')
        super().__init__(file_name=file_name, folder_path=folder_path, **kwargs)
        self.metadata_file = kwargs.get("metadata_file", "metadata.json.gz")

    def set_full_path(self, nested_dirs: list = None):
        """Set full path to write to"""
        # creates full path
        full_path = [self.folder_path]+nested_dirs
        return os.path.join(*full_path).replace('\\', '/')

    # pylint: disable=no-member,too-many-arguments
    def write_to_s3(
            self,
            json_data,
            data_path,
            mode="w",
            encoding="utf-8",
            ensure_ascii=False,
            default=None
    ):
        """
            Write report to s3 bucket
        """
        buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode=mode) as file_:
            with io.TextIOWrapper(file_, encoding=encoding) as textio_obj:
                textio_obj.write(json.dumps(json_data, ensure_ascii=ensure_ascii, default=default))
        # write to s3
        buffer.seek(0)
        self.s3_resource.Bucket(self.bucket).put_object(
            Key=os.path.join(self.full_path, data_path).replace('\\', '/'),
            Body=buffer)

    def write_metadata(self, metadata: dict):
        """"
            Creates required GA/ViewID/metricName structure
        """
        # create directory structure if not exists
        self.full_path = self.set_full_path(nested_dirs=[
            metadata.get("viewID", '.'),
            metadata.get("metricName", '')])
        self.write_to_s3(json_data=metadata, data_path=self.metadata_file, mode="w")

    def write_file(self):
        """
            upload python dict into s3 bucket with gzip archive
        """
        # generate a unix timestamp suffix for file
        suff = self.timestamp_suffix()
        data_path = "{}_{}.json.gz".format(self.file_name, suff)
        self.write_to_s3(self.data, data_path, mode="wb")
        self.data = []



"""
    utils functions
"""
import yaml


def init_config(file_name: str = './example.yaml'):
    """
        Util method that creates an example config at the given file_name
        args:
            file_name (str): full path to config YAML file
    """
    config = {
        'queryConfig': [{
            'endpoint': 'stat',
            'parameters':[
                {'stats': 'list',
                 'list': 'alert subscribers'}]}]}

    # write to file
    with open(file_name, 'w') as file_:
        yaml.dump(config, file_)



"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import os
from typing import Callable

import httplib2
from googleapiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials

from sdc_dp_helpers.api_utilities.retry_managers import request_handler

class BaseGAV4Reader:
    """
        Base Google Analytics V4 reader class.

        Used as basis for building custom GA V4 readers, it requires that you
        add your own Authentication through the auth_credentials method.
        The basic workflow is:
        (1) Create credentials
        (2) Build google service object
        (3) Return an iterable of the get_report method
        (4) Iterate until no pagination token is found. A.k.a. pageToken==None
    """
    analytics = None
    max_pages = None

    class ReaderIter:
        """Inner Class - Iterator that yields queries until max pages or pageToken == None."""
        page_counter = 0

        def __init__(
                self,
                *,
                query_fn: Callable,
                query: list,
                page_token: str = None,
                **kwargs
        ):
            self.query_fn = query_fn
            self.max_pages = kwargs.get("max_pages", None)
            self.page_token = page_token
            self.query = query

        def __iter__(self):
            """Iterator method for iterator class"""
            return self

        def __next__(self):
            """
                Next method for iterator class which includes
                the StopIteration criteria for iterable
            """
            # check StopIteration criteria
            if self.max_pages is not None:
                if self.page_counter >= self.max_pages:
                    print("stopping iteration on max pages = {}".format(self.max_pages))
                    raise StopIteration

            if self.page_counter > 0 and self.page_token is None:
                print("stopping iteration on pageToken = {}".format(self.page_token))
                raise StopIteration

            # run get_report
            report, self.page_token = self.query_fn(
                queries=self.query,
                page_token=self.page_token
            )
            print(
                'Finished query function for page {page} of view ID {view_id}'.format(
                    page=self.page_counter + 1,
                    view_id=self.query[0].get('viewId')
                )
            )

            # increment pagination counter
            self.page_counter += 1

            return report

    # pylint: disable=no-member
    def build_analytics_service_object(
            self,
            api_name: str = 'analyticsreporting',
            api_version: str = 'v4'
    ):
        """Initializes the analytics reporting service object.
        args:
        client_secrets_path (json) : Google API OAuth2 API credentials
        scopes (list or iterable) : scopes for Google OAuth2 authentication

        Returns:
        analytics an authorized analyticsreporting service object.
        """
        # authorize HTTP object
        http = self.credentials.authorize(http=httplib2.Http())
        # Build the service object.
        return build(api_name, api_version, http=http)

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 0.01)
    )
    def get_report(
            self,
            queries: list,
            page_token: str = None
    ):
        """
            Use the Analytics Service Object to query the Analytics Reporting API V4.

            args:
            analytics : analytics reporting service object
            query (list): array of queries. [max. 5]
            page_token (str): token sent denoting the last scraped data point
            return:
            analytics report data (json)

        """
        if page_token is not None:
            for query in queries:
                query.update({'page_token':page_token})

        # run batch query
        report = self.analytics.reports().batchGet(body={
            'reportRequests': queries}).execute().get('reports')
        page_token = report[0].get('nextPageToken')

        return report, page_token

    def run_query(self, query: list, page_token: str = None):
        """
            returns a sync iterator
        """
        iter_ = self.ReaderIter(
            query_fn=self.get_report,
            query=query,
            max_pages=self.max_pages,
            page_token=page_token)
        return iter_


class CustomGoogleAnalyticsReaderWithServiceAcc(BaseGAV4Reader):
    """
        Custom Google Analytics Reader which authenticates with Service account
    """

    # pylint: disable=super-init-not-called
    def __init__(
            self,
            service_account_secrets_path: str,
            service_account_email: str,
            scopes: list,
            **kwargs
    ):
        self.max_pages = kwargs.get("max_pages", None)
        self.credentials = self.auth_credentials(
            service_account_secrets_path, service_account_email, scopes)
        # build analytics object
        self.analytics = self.build_analytics_service_object()

    @staticmethod
    def auth_credentials(
            service_account_secrets_path,
            service_account_email,
            scopes
    ):
        """Get a service that communicates to a Google API.
        Args:
            service_account_secrets_path: The filepath to service secrets.
            api_name: The name of the api to connect to.
            api_version: The api version to connect to.
            scope: A list auth scopes to authorize for the application.
            key_file_location: The path to a valid service account p12 key file.
            service_account_email: The service account email address.
            scopes: Allowed scopes for the given credentials
        Returns:
            A service that is connected to the specified API.
        """
        return ServiceAccountCredentials.from_p12_keyfile(
            service_account_email, service_account_secrets_path, scopes=scopes)



"""
    CONFIG UTILITIES
        - classes that parse your config file, build a query iterable and yield queries
"""
import yaml


# pylint: disable=invalid-name, unused-argument
def _generate_queries(query_config: list):
    """
    args:
        queryConfig (list): SailThru query config containing endpoint, parameters, etc.
    yields:
        query generator of queries to run
    """
    for config in query_config:
        endpoint = config.get('endpoint', None)
        if endpoint is None:
            raise Exception('ClientError: Expected parameter endpoint.')

        parameters = config.get('parameters', None)
        if endpoint is None:
            raise Exception('ClientError: Expected parameters.')

        fields = {
            "endpoint": endpoint,
            "parameters": parameters}
        # building similar queries per vid
        yield [fields]


def _parser(file_name: str) -> list:
    """Parse config in to dictionary"""
    # load config
    with open(file_name, 'r') as file_:
        config = yaml.safe_load(file_)

    return config.get('queryConfig')

# pylint: disable=too-few-public-methods
class SailThruConfigManager:
    """
        Class that parses config .yaml for queries, reader, writer
        Uses these configs to build required query objects.
    """

    def __init__(self, file_name):
        # load config
        self.config = _parser(file_name=file_name)

    def get_query(self, config=None):
        """
        Gets the query parameters from the config file
        :param config: configurations
        :return: query parameters
        """
        if config is None:
            config = self.config

        return _generate_queries(config)
